; 
    include \masm32\include\masm32rt.inc
; 
.686p

comment * -----------------------------------------------------
                     Build this console app with
                  "MAKEIT.BAT" on the PROJECT menu.
        ----------------------------------------------------- *

    .data
    ;;;;;;; CONFIGURATION ;;;;;;;
    ;; gu_* stands for global user preference. These are meant to be adjusted
    ;; independently of any other code.
      ; What priority should the test driver run at?
      gu_priorityClass textequ <REALTIME_PRIORITY_CLASS>

      ; Number of successive runs through the test code that must take the
      ; exact same number of CPU ticks. This setting allows us to drop
      ; spurious results like context switches, interrupts, CPU throttling,
      ; cache misses and whatever other "one-off" events occur.
      ; NOTE: this number will be automatically adjusted by the test driver
      ; on an as needed basis to ensure test results are stable and timely.
      ; This is essentially the starting point for the driver.
      gu_default_sample_batch_size equ 16

      ; The driver may raise the sample_batch_size no higher than this number.
      ; Some combinations of opcodes are super stable and give spurious results
      ; at lower settings, so the driver will attempt to adjust the testing to
      ; the code sequence under test. (Currently UNIMPLEMENTED)
      gu_max_sample_batch_size equ 512

      ; How many times do we attempt to get a stable sample reading before
      ; giving up and adjusting downwards the sample_batch_size?
      gu_sample_retry_attempts equ 100000
     
      ; Number of fully stable samples we take and toss away.
      ; We do this as an attempt to stabalize the cpu at full throttle.
      ; Play with this number, the lower you take it, the less lagtime
      ; you will have before results start getting printed out. The downside
      ; is the initial numbers may not be statistically relevant.
      ; Remember the ideal situtation is to only emit numbers that are the same
      ; from one run to the next! That is, we want a standard deviation of 0!
      gu_warmup_samples equ 1 ;600

      ; Total number of samples to print out. In "production" versions
      ; this ought to be some small number like 8 or 16.
      gu_total_samples equ 100
    ;;;;;;; END CONFIG ;;;;;;;;;;
    

      align 16
      g_run_cache dd gu_max_sample_batch_size dup(0)
      align 16
      g_steps dd gu_default_sample_batch_size-1
      align 16
      g_result_timings dd gu_total_samples dup(0)

      align 16
      gu_testfunction dd testit
    .code

start:
   
; 
    invoke GetCurrentProcess
    invoke SetPriorityClass, eax, gu_priorityClass
    invoke GetCurrentThread
    invoke SetThreadPriority, eax, THREAD_PRIORITY_TIME_CRITICAL
    invoke  GetCurrentProcess
    invoke  SetProcessAffinityMask,eax,1
    invoke  Sleep,750
    invoke Sleep, 0
    call main
   ; inkey
    exit

; 

main proc
    mov ecx, 31
    l_spinup:
      call init_spinup
      dec ecx
    jns l_spinup
    
    cls
    ; take some samples, but toss the results away.
    ; we do this to spin the selected core up.
    mov ecx, gu_warmup_samples-1
    ;test ecx,0
    ;jns l_end_prep
    l_prep:

    align 16
    push ecx
    call do_sample_run
    pop ecx
    dec ecx
    jns l_prep
    l_end_prep:
    ; now run for real, printing each result to standard out.
    mov ecx, gu_total_samples-1
    align 16
l:
    call do_sample_run
    ;push eax
    mov [g_result_timings+ecx*4], eax
    ;call print_number
    dec ecx
    jns l

    mov ecx, gu_total_samples-1
l_final_print:
    mov eax, [g_result_timings+ecx*4]
    push eax
    call print_number
    dec ecx
    jns l_final_print
    
    ret
main endp

;; take samples until the cpu takes the same amount of time to do N runs.
;; RETURN: EAX=ticks taken.
do_sample_run proc
    push ecx
    push ebp

reset_and_retry:
    mov ebp, [g_steps]
    mov ecx, gu_sample_retry_attempts
retry:
    ;invoke Sleep,0
    dec ecx
    jz stepdown
    push ecx
    call spinup

    mov edi, ebp
l:
    push edi
    call timeit
    mov [g_run_cache+edi*4], eax
    pop edi
    dec edi
    jns l

    mov edi, ebp
    mov eax, [g_run_cache+edi*4]
    pop ecx
l_check_equality:
    cmp [g_run_cache+edi*4], eax
    jne retry
    dec edi
    jns l_check_equality

    pop ebp
    pop ecx
    ret
 stepdown:
    sub [g_steps], 1
    ;print "Now requiring only " ; need a better name for this... xD
    ;mov eax, [g_steps]
    ;add eax, 1
    ;print ustr$(eax)
    ;print " entries.",13,10
    jmp reset_and_retry
do_sample_run endp

;; Does not visibly change any registers.
;; Simply spins up the CPU in preperation of "real" timing tests.
align 16
init_spinup proc
    push ecx
    mov ecx, 10000000
l:
    dec ecx
    jns l
    pop ecx
    ret
init_spinup endp


;; Does not visibly change any registers.
;; Simply spins up the CPU in preperation of "real" timing tests.
align 16
spinup proc
    push ecx
    mov ecx, 100000
l:
    dec ecx
    jns l
    pop ecx
    ret
spinup endp



align 16
timeit proc
; warmup
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    xor eax, eax
    cpuid
    ;rdtsc 
    ; now go for real!
    xor eax, eax
    cpuid
    rdtsc
   
    
    push eax
    push edx
    
    
    call [gu_testfunction]



    xor eax, eax
    cpuid
    rdtsc
    push eax
    push edx
    call print_runtime

    ret

timeit endp


print_runtime proc
    pop esi
    pop edx
    pop eax
    pop ecx

    cmp ecx, edx ; if RDTSC wrapped around, the high 32 bits will not be equal.
    jne wraparound
    
    pop ecx
    sub eax, ecx
    push esi
    ret
    
    align 16
wraparound:
    mov eax, 0
    pop ecx
    push esi
    ret
print_runtime endp

;; destroys the eax, edx register, not used for return result.
align 16
print_number  proc
    pop ebp
    pop eax
    push ecx
    print ustr$(eax)
    print ", ";,13,10
    pop ecx
    push ebp
    ret
print_number endp

calc_minimum proc

calc_minimum endp

align 16

testit_baseline proc
    mov ecx, 15
  l:
    dec ecx
    jns l
    ret
testit_baseline endp

align 16
;;;;;;;;; code under test should go here! ;;;;;;;;;
;; eax, ecx, edx, ebx are free scratch registers
testit proc
 mov ecx, 15
 l:
 
   ;
   ; imul eax, ecx, 1000
   ; imul ebx, eax, 10000
    ;pop ecx
    dec ecx
    jns l
   
ret
testit endp




; 

end start
